%% Copyright (C) 2024 by sysmocom - s.f.m.c. GmbH <info@sysmocom.de>
%% Author: Vadim Yanitskiy <vyanitskiy@sysmocom.de>
%%
%% All Rights Reserved
%%
%% SPDX-License-Identifier: AGPL-3.0-or-later
%%
%% This program is free software; you can redistribute it and/or modify
%% it under the terms of the GNU Affero General Public License as
%% published by the Free Software Foundation; either version 3 of the
%% License, or (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU Affero General Public License
%% along with this program.  If not, see <https://www.gnu.org/licenses/>.
%%
%% Additional Permission under GNU AGPL version 3 section 7:
%%
%% If you modify this Program, or any covered work, by linking or
%% combining it with runtime libraries of Erlang/OTP as released by
%% Ericsson on https://www.erlang.org (or a modified version of these
%% libraries), containing parts covered by the terms of the Erlang Public
%% License (https://www.erlang.org/EPLICENSE), the licensors of this
%% Program grant you additional permission to convey the resulting work
%% without the need to license the runtime libraries of Erlang/OTP under
%% the GNU Affero General Public License. Corresponding Source for a
%% non-source form of such a combination shall include the source code
%% for the parts of the runtime libraries of Erlang/OTP used as well as
%% that of the covered work.

-module(sctp_proxy).
-behaviour(gen_statem).

-export([init/1,
         callback_mode/0,
         connecting/3,
         connected/3,
         code_change/4,
         terminate/3]).
-export([start_link/3,
         send_data/2,
         shutdown/1]).

-include_lib("kernel/include/logger.hrl").
-include_lib("kernel/include/inet.hrl").
-include_lib("kernel/include/inet_sctp.hrl").

%% ------------------------------------------------------------------
%% public API
%% ------------------------------------------------------------------

start_link(Aid, MmeAddr, MmePort) ->
    gen_statem:start_link(?MODULE, [Aid, MmeAddr, MmePort], []).


send_data(Pid, Data) ->
    gen_statem:cast(Pid, {send_data, Data}).


shutdown(Pid) ->
    gen_statem:stop(Pid).


%% ------------------------------------------------------------------
%% gen_statem API
%% ------------------------------------------------------------------

init([Aid, MmeAddr, MmePort]) ->
    {ok, connecting,
     #{enb_aid => Aid,
       mme_addr => MmeAddr,
       mmr_port => MmePort,
       tx_queue => []}}.


callback_mode() ->
    [state_functions, state_enter].


%% CONNECTING state
connecting(enter, OldState,
           #{mme_addr := MmeAddr, mmr_port := MmePort} = S) ->
    ?LOG_INFO("State change: ~p -> ~p", [OldState, ?FUNCTION_NAME]),
    %% Initiate connection establishment with the MME
    {ok, Sock} = sctp_client:connect(MmeAddr, MmePort),
    {keep_state, S#{sock => Sock},
     [{state_timeout, 2_000_000, conn_est_timeout}]};

%% Handle connection establishment timeout
connecting(state_timeout, conn_est_timeout, _S) ->
    {stop, {shutdown, conn_est_timeout}};

%% Handle an eNB -> MME data forwarding request (queue)
connecting(cast, {send_data, Data},
           #{tx_queue := Pending} = S) ->
    {keep_state, S#{tx_queue := [Data | Pending]}};

%% Handle an #sctp_assoc_change event (connection state)
connecting(info, {sctp, _Socket, MmeAddr, MmePort,
                  {[], #sctp_assoc_change{state = ConnState,
                                          assoc_id = Aid}}}, S) ->
    case ConnState of
        comm_up ->
            ?LOG_NOTICE("MME connection (id=~p, ~p:~p) established",
                        [Aid, MmeAddr, MmePort]),
            {next_state, connected, S#{mme_aid => Aid}};
        _ ->
            ?LOG_NOTICE("MME connection establishment failed: ~p", [ConnState]),
            {stop, {shutdown, conn_est_fail}}
    end;

%% Catch-all for other kinds of SCTP events
connecting(info, {sctp, _Socket, MmeAddr, MmePort,
                  {AncData, Data}}, S) ->
    ?LOG_DEBUG("Unhandled SCTP event (~p:~p): ~p, ~p",
               [MmeAddr, MmePort, AncData, Data]),
    {keep_state, S};

connecting(Event, EventData, S) ->
    ?LOG_ERROR("Unexpected event ~p: ~p", [Event, EventData]),
    {keep_state, S}.


%% CONNECTED state
connected(enter, OldState, S) ->
    ?LOG_INFO("State change: ~p -> ~p", [OldState, ?FUNCTION_NAME]),
    %% Send pending eNB -> MME messages (if any)
    ok = sctp_send_pending(S),
    {keep_state, maps:remove(tx_queue, S)};

%% Handle an eNB -> MME data forwarding request (forward)
connected(cast, {send_data, Data}, S) ->
    ok = sctp_send(S, Data),
    {keep_state, S};

%% Handle an #sctp_assoc_change event (MME connection state)
connected(info, {sctp, _Socket, MmeAddr, MmePort,
                 {[], #sctp_assoc_change{state = ConnState,
                                         assoc_id = Aid}}}, S) ->
    case ConnState of
        comm_up ->
            ?LOG_NOTICE("MME connection (id=~p, ~p:~p) is already established?!?",
                        [Aid, MmeAddr, MmePort]),
            {keep_state, S};
        _ ->
            ?LOG_NOTICE("MME connection state: ~p", [ConnState]),
            {stop, {shutdown, conn_fail}}
    end;

%% Handle an #sctp_sndrcvinfo event (MME -> eNB data)
connected(info, {sctp, _Socket, MmeAddr, MmePort,
                 {[#sctp_sndrcvinfo{assoc_id = Aid}], Data}}, S) ->
    ?LOG_DEBUG("MME connection (id=~p, ~p:~p) -> eNB: ~p",
               [Aid, MmeAddr, MmePort, Data]),
    sctp_server:send_data(maps:get(enb_aid, S), handle_pdu(Data)),
    {keep_state, S};

%% Catch-all for other kinds of SCTP events
connected(info, {sctp, _Socket, MmeAddr, MmePort,
                 {AncData, Data}}, S) ->
    ?LOG_DEBUG("Unhandled SCTP event (~p:~p): ~p, ~p",
               [MmeAddr, MmePort, AncData, Data]),
    {keep_state, S};

%% Catch-all handler for this state
connected(Event, EventData, S) ->
    ?LOG_ERROR("Unexpected event ~p: ~p", [Event, EventData]),
    {keep_state, S}.



code_change(_Vsn, State, S, _Extra) ->
    {ok, State, S}.


terminate(Reason, State, S) ->
    ?LOG_NOTICE("Terminating in state ~p, reason ~p", [State, Reason]),
    case S of
        #{sock := Sock, mme_aid := Aid} ->
            sctp_client:disconnect({Sock, Aid}),
            gen_sctp:close(Sock);
        #{sock := Sock} ->
            gen_sctp:close(Sock);
        _ -> ok
    end.


%% ------------------------------------------------------------------
%% private API
%% ------------------------------------------------------------------

%% A safe wrapper for s1ap_proxy:handle_pdu/1
-spec handle_pdu(binary()) -> binary().
handle_pdu(Data) when is_binary(Data) ->
    try s1ap_proxy:handle_pdu(Data) of
        NewData -> NewData
    catch
        Exception:Reason ->
            ?LOG_ERROR("An exception occurred: ~p, ~p", [Exception, Reason]),
            Data %% proxy as-is
    end.


%% Send a single message to the MME
sctp_send(#{sock := Sock, mme_aid := Aid}, Data) ->
    sctp_client:send_data({Sock, Aid}, handle_pdu(Data)).


%% Send pending messages to the MME
sctp_send_pending(#{tx_queue := Pending} = S) ->
    sctp_send_pending(S, lists:reverse(Pending)).

sctp_send_pending(S, [Data | Pending]) ->
    sctp_send(S, Data),
    sctp_send_pending(S, Pending);

sctp_send_pending(_S, []) ->
    ok.

%% vim:set ts=4 sw=4 et:
